/* $Header: https://svn.mpia.de/gulli/TwiceAsNice/branch/unstable/TwiceAsNice/nice/src/libNice/Nice/Alg/Alg.cc 15141 2018-02-21 14:14:57Z mathar $
*/

#include <Nice/Alg/Alg.h>

/**
* @brief linear least-square solver using normal equations
*
* Given a set of data points _x[0..ndat-1], _y[0..ndat-1] with inidividual
* standard deviations sig[0..ndat-1], use chi-squared minimization to fit for some
* or all of the coefficeints a[0..ma-1] of a function that depends linearly on a,
* y=sum_i a_i*afunc_i(x). The boolean input array ia[0..ma-1] indicates by
* true entries those components of a that should be fitted for, and by false
* entries those components that should be held fixed at their input values.
* The program returns values for a[0..ma-1], chi-square=chisq, and the covariance
* matrix covar[0..ma-1][0..ma-1]. (Parameters held fixed will return zero
* covariances.) The user supplies a routine funcs(x,afunc) that returns the
* ma basis functions evaluated at x=x in the array afunc[0..ma-1]).
* (Excerpt of numerical recipies section 15.4)
* @param[in] _x     data points (x)
* @param[in] _y     data points (y)
* @param[in] _sig   standard deviations of the data points
* @param[out] _a     coefficients
* @param[in] _ia    flags for each coefficient. if false, coefficient will be kept constant
* @param[out] _covar covariance matrix will be returned
* @param[out] _chisq chi^2 as return value
* @param[in] _funcs the function that is linear to the coefficients a
*/
void Nice::Alg::linearFit ( boost::numeric::ublas::vector<double> &_x,
                            boost::numeric::ublas::vector<double> &_y,
                            boost::numeric::ublas::vector<double> &_sig,
                            boost::numeric::ublas::vector<double> &_a,
                            boost::numeric::ublas::vector<bool> &_ia,
                            boost::numeric::ublas::matrix<double> &_covar,
                            double &_chisq,
                            void _funcs ( const double, boost::numeric::ublas::vector<double> & ) )
{
    int i,j,k,l,m,mfit=0;
    double ym,wt,sum,sig2i;

    int ndat=(int)_x.size();
    int ma=(int)_a.size();
    boost::numeric::ublas::vector<double> afunc ( ma );
    boost::numeric::ublas::matrix<double> beta ( ma,1 );
    for ( j=0;j<ma;j++ )
        if ( _ia[j] ) mfit++;

    if ( mfit == 0 )
    {
        std::string error ( "Fit parameter are missing." );
        E_LOG ( error );
        E_THROW ( ::Nice::Algorithm::NoFitParamsException, error );
    }

    for ( j=0;j<mfit;j++ )
    {
        for ( k=0;k<mfit;k++ ) _covar ( j,k ) =0.0;
        beta ( j,0 ) =0.0;
    }
    for ( i=0;i<ndat;i++ )
    {
        _funcs ( _x[i],afunc );
        ym=_y[i];
        if ( mfit < ma )
        {
            for ( j=0;j<ma;j++ )
                if ( !_ia[j] ) ym -= _a[j]*afunc[j];
        }
        sig2i=1.0/SQR ( _sig[i] );
        for ( j=0,l=0;l<ma;l++ )
        {
            if ( _ia[l] )
            {
                wt=afunc[l]*sig2i;
                for ( k=0,m=0;m<=l;m++ )
                    if ( _ia[m] ) _covar ( j,k++ ) += wt*afunc[m];
                beta ( j++,0 ) += ym*wt;
            }
        }
    }
    for ( j=1;j<mfit;j++ )

        for ( k=0;k<j;k++ )
            _covar ( k,j ) =_covar ( j,k );
    boost::numeric::ublas::matrix<double> temp ( mfit,mfit );
    for ( j=0;j<mfit;j++ )
        for ( k=0;k<mfit;k++ )
            temp ( j,k ) =_covar ( j,k );
    gaussj ( temp,beta );
    for ( j=0;j<mfit;j++ )
        for ( k=0;k<mfit;k++ )
            _covar ( j,k ) =temp ( j,k );
    for ( j=0,l=0;l<ma;l++ )
        if ( _ia[l] ) _a[l]=beta ( j++,0 );
    _chisq=0.0;
    for ( i=0;i<ndat;i++ )
    {
        _funcs ( _x[i],afunc );
        sum=0.0;
        for ( j=0;j<ma;j++ ) sum += _a[j]*afunc[j];
        _chisq += SQR ( ( _y[i]-sum ) /_sig[i] );
    }
    covsrt ( _covar,_ia,mfit );
} /* linearFit */

/** @brief linear least-square solver using normal equations
  *
  * Given a set of data points _x[0..ndat-1], _y[0..ndat-1] with inidividual
  * standard deviations sig[0..ndat-1], use chi-squared minimization to fit for some
  * or all of the coefficeints a[0..ma-1] of a function that depends linearly on a,
  * y=sum_i a_i*afunc_i(x). The boolean input array ia[0..ma-1] indicates by
  * true entries those components of a that should be fitted for, and by false
  * entries those components that should be held fixed at their input values.
  * The program returns values for a[0..ma-1], chi-square=chisq, and the covariance
  * matrix covar[0..ma-1][0..ma-1]. (Parameters held fixed will return zero
  * covariances.) The user supplies a routine funcs(x,afunc) that returns the
  * ma basis functions evaluated at x=x in the array afunc[0..ma-1]).
  * (Excerpt of numerical recipies section 15.4)
  * @param[in] _x     data points (x)
  * @param[in] _y     data points (y)
  * @param[in] _sig   standard deviations of the data points
  * @param[out] coeffs     coefficients
  * @param[in] fitCoeffs    flags for each coefficient. if false, coefficient will be kept constant
  * @param[out] _covar covariance matrix will be returned
  * @param[out] _chisq chi^2 as return value
  * @param[in] _funcs the function that is linear to the coefficients a */
void Nice::Alg::shinsLinearFit(boost::numeric::ublas::vector<double> &_x,
                               boost::numeric::ublas::vector<double> &_y,
                               boost::numeric::ublas::vector<double> &_sig,
                               boost::numeric::ublas::vector<double> &coeffs,
                               boost::numeric::ublas::vector<bool> &fitCoeffs,
                               boost::numeric::ublas::matrix<double> &_covar,
                               double &_chisq,
                               void base_poly_3 ( const double, boost::numeric::ublas::vector<double> & ) )
{
    int i, j, k, l, m, mfit = 0;
    double ym, wt, sum, sig2i;

    int num_dat = (int)_x.size();
    int num_coeff = (int)coeffs.size();
    boost::numeric::ublas::vector<double> afunc(num_coeff);
    boost::numeric::ublas::matrix<double> beta(num_coeff, 1);
    for(j = 0; j < num_coeff; j++)
    {
        if(fitCoeffs.size() > j && fitCoeffs[j])mfit++;
    }

    if(mfit == 0 || _y.size() != num_dat)
    {
        std::string error;
        if(mfit == 0)error = "Fit parameter are missing.";
        else if(_x.size() != _y.size())error = "Vector X ad Y must have the same size.";
        E_LOG ( error );
        E_THROW ( ::Nice::Algorithm::NoFitParamsException, error );
    }

    for(j = 0; j < mfit; j++)
    {
        for(k = 0; k < mfit; k++)
        {
            _covar(j, k) = 0.0;
        }
        beta(j, 0) = 0.0;
    }

    for(i = 0; i < num_dat; i++)
    {
        base_poly_3(_x[i], afunc);
        ym = _y[i];
        if(mfit < num_coeff)
        {
            for(j = 0; j < num_coeff; j++)
            {
                if(!fitCoeffs[j])
                    ym -= coeffs[j]*afunc[j];
            }
        }
        sig2i = 1.0/SQR(_sig[i]);
        for(j = 0, l = 0; l < num_coeff; l++)
        {
            if(fitCoeffs[l])
            {
                wt = afunc[l]*sig2i;
                for(k = 0 , m = 0; m <= l; m++)
                {
                    if(fitCoeffs[m])
                        _covar(j, k++) += wt*afunc[m];
                }
                beta(j++, 0) += ym*wt;
            }
        }
    }
    for(j = 1; j < mfit; j++)
    {
        for(k = 0; k < j; k++)
        {
            _covar(k, j) = _covar (j, k);
        }
    }
    boost::numeric::ublas::matrix<double> temp(mfit, mfit);
    for(j = 0; j < mfit; j++)
    {
        for(k = 0; k < mfit; k++)
        {
            temp(j, k) = _covar(j, k);
        }
    }
    gaussj(temp, beta);
    for(j = 0; j < mfit; j++)
    {
        for(k = 0; k < mfit; k++)
        {
            _covar( j, k) = temp(j, k);
        }
    }
    for(j = 0, l = 0; l < num_coeff; l++)
    {
        if(fitCoeffs[l])
            coeffs[l] = beta(j++, 0);
    }
    _chisq = 0.0;
    for(i = 0; i < num_dat; i++)
    {
        base_poly_3(_x[i], afunc);
        sum = 0.0;
        for(j = 0; j < num_coeff; j++)
        {
            sum += coeffs[j]*afunc[j];
        }
        _chisq += SQR((_y[i] - sum)/_sig[i]);
    }
    covsrt(_covar, fitCoeffs, mfit);
} /* shinsLinearFit */



/**
* @brief Gauss-Jordan elimination
*
* Used e.g. by lfit to solve for the coefficients (and to find the covariance matrix)
*
* @param _a
* @param _b
*/
void Nice::Alg::gaussj ( boost::numeric::ublas::matrix<double> &a, boost::numeric::ublas::matrix<double> &b )
{
    int i,icol=0,irow=0,j,k,l,ll;
    double big,dum,pivinv;

    int n=(int)a.size1();
    int m=(int)b.size2();
    boost::numeric::ublas::vector<int> indxc ( n ),indxr ( n ),ipiv ( n );
    for ( j=0;j<n;j++ ) ipiv[j]=0;
    for ( i=0;i<n;i++ )
    {
        big=0.0;
        for ( j=0;j<n;j++ )
            if ( ipiv[j] != 1 )
                for ( k=0;k<n;k++ )
                {
                    if ( ipiv[k] == 0 )
                    {
                        if ( fabs ( a ( j,k ) ) >= big )
                        {
                            big=fabs ( a ( j,k ) );
                            irow=j;
                            icol=k;
                        }
                    }
                }
        ++ ( ipiv[icol] );
        if ( irow != icol )
        {
            for ( l=0;l<n;l++ ) SWAP ( a ( irow,l ),a ( icol,l ) );
            for ( l=0;l<m;l++ ) SWAP ( b ( irow,l ),b ( icol,l ) );
        }
        indxr[i]=irow;
        indxc[i]=icol;

        if ( a ( icol,icol ) == 0.0 )
        {
            std::string error ( "gauss j singular matrix exception." );
            E_LOG ( error );
            E_THROW ( ::Nice::Algorithm::GaussjSingularMatrixException, error );
        }

        pivinv=1.0/a ( icol,icol );
        a ( icol,icol ) =1.0;
        for ( l=0;l<n;l++ ) a ( icol,l ) *= pivinv;
        for ( l=0;l<m;l++ ) b ( icol,l ) *= pivinv;
        for ( ll=0;ll<n;ll++ )
            if ( ll != icol )
            {
                dum=a ( ll,icol );
                a ( ll,icol ) =0.0;
                for ( l=0;l<n;l++ ) a ( ll,l ) -= a ( icol,l ) *dum;
                for ( l=0;l<m;l++ ) b ( ll,l ) -= b ( icol,l ) *dum;
            }
    }
    for ( l=n-1;l>=0;l-- )
    {
        if ( indxr[l] != indxc[l] )
            for ( k=0;k<n;k++ )
                SWAP ( a ( k,indxr[l] ),a ( k,indxc[l] ) );
    }
} /* gaussj */

/**
* @brief Sorts covariance matrix
*
* @param _covar
* @param _ia
* @param _mfit
*/
void Nice::Alg::covsrt ( boost::numeric::ublas::matrix<double> &covar, boost::numeric::ublas::vector<bool> &ia, const int mfit )
{
    int i,j,k;

    int ma=(int)ia.size();
    for ( i=mfit;i<ma;i++ )
        for ( j=0;j<i+1;j++ ) covar ( i,j ) =covar ( j,i ) =0.0;
    k=mfit-1;
    for ( j=ma-1;j>=0;j-- )
    {
        if ( ia[j] )
        {
            for ( i=0;i<ma;i++ ) SWAP ( covar ( i,k ),covar ( i,j ) );
            for ( i=0;i<ma;i++ ) SWAP ( covar ( k,i ),covar ( j,i ) );
            k--;
        }
    }
} /* covsrt */


#ifndef M_SQRT3_2
/** Altitude in a regular triangle of unit side length.
* sqrt(3)/2 = 0.5*1.7320508075.
*/
#define M_SQRT3_2 0.866025403784438646763723170755
#endif

/*!******************************************
* @brief construct a chain of points on a c.c.w. spiral.
*
* @param c A circle which will contain the spiral.
*   The center of the circle defines the starting point of the spiral.
*   The radius of the circle defines how far the spiral radius may reach out from the center.
* @param type Type of the discret underlying lattice.
*   If type=0 the points are on a simple 2D quadratic lattice.
*   If type=1 the points are on a simple triangular (a.k.a. hexagonal) lattice.
*   There is no type for centered square lattices, because these are the
*   same as quadratic lattices rotated by 45 degrees and shrunk by a factor sqrt(2).
*   (So they may be derived by rotation and shrinking the vector of points returned by this function.)
* @param latConst The lattice constant in the same units as the circle center and radius.
*   This is the shortest distance between two distinct points on the lattice.
*   That value should not be an integer multiple of the lattice constant (to avoid fuzzy interpretation
*   of which points on the spiral stay within the circle)
* @return A sequence of points with increasing distance from the circle center.
* @since 2017-11-02
* @author R. J. Mathar
*/
Nice::SeqPoint Nice::Alg::generateSpiral(const Nice::Circle & c, int type, double latConst)
{
	/* the list of all points to be returned */
	Nice::SeqPoint spir ;

	/* the current tail of the list */
	Nice::Point tail(c.getCenter()) ;

	/* the current heading of the spiral's end, starting horizontal to +x */
	Nice::Point delta(latConst,0.) ;

	switch (type)
	{
	case 0:
		spir.push_back(tail) ;
		/* each generation means walking along the perimeter of a square
		*/
		for(int gen =1;; gen++)
		{
			/* distances will all be beyond the circle radius: stop iteration
			*/
			if ( gen*latConst > c.r)
				break ;

			/* one step to the right */
			tail += delta ;
			spir.push_back(tail) ;

			/* 2*gen-1 steps up */
			delta = Nice::Point(0.,latConst) ;
			for(int s = 0; s < 2*gen-1 ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}

			/* 2*gen steps left */
			delta = Nice::Point(-latConst,0.) ;
			for(int s = 0; s < 2*gen ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}

			/* 2*gen steps down */
			delta = Nice::Point(0.,-latConst) ;
			for(int s = 0; s < 2*gen ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}

			/* 2*gen steps right */
			delta = Nice::Point(latConst,0.) ;
			for(int s = 0; s < 2*gen ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}
		}
		break ;
	case 1:
		spir.push_back(tail) ;
		/* each generation means walking along the perimeter of a regular hexagon
		* where the furthest distance on the hexagon (occuring 6 times) to the center is gen*latConst.
		*/
		for(int gen =1;; gen++)
		{
			/* distances will all be beyond the circle radius: stop iteration
			* Here the left hand side is the distance to any of the 6 sides of the hexagon.
			*/
			if ( M_SQRT3_2* gen*latConst > c.r)
				break ;

			/* one step to the right */
			tail += delta ;
			spir.push_back(tail) ;

			/* gen-1 steps right up */
			delta = Nice::Point(0.5*latConst,M_SQRT3_2*latConst) ;
			for(int s = 0; s < gen-1 ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}

			/* gen steps left up */
			delta = Nice::Point(-0.5*latConst,M_SQRT3_2*latConst) ;
			for(int s = 0; s < gen ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}

			/* gen steps left */
			delta = Nice::Point(-latConst,0.) ;
			for(int s = 0; s < gen ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}

			/* gen steps left down */
			delta = Nice::Point(-0.5*latConst,-M_SQRT3_2*latConst) ;
			for(int s = 0; s < gen ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}

			/* gen steps right down */
			delta = Nice::Point(0.5*latConst,-M_SQRT3_2*latConst) ;
			for(int s = 0; s < gen ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}

			/* gen steps right */
			delta = Nice::Point(latConst,0.) ;
			for(int s = 0; s < gen ; s++)
			{
				tail += delta ;
				if ( c.contains(tail)) 
					spir.push_back(tail) ;
			}
		}
		break ;
	}
	return spir ;
} /* generateSpiral */

/*!******************************************
* @brief construct subsets of size m out of n elements
*
* This is the standard combinatorial sub-set construction.
* For N=5 , m=3 we construct <0,1,2>, <0,1,3>, <0,1,4>,
*  <0,2,3>, <0,2,4>...., <2,3,4>.
*
* @param N The number of elements to choose from.
* @param m The number of elements selected in each element of the result.
* @return ordered lists of numbers in the range 0..n-1 where
*  each list has m elements.
*  The cardinality of the outer vector is biomial(N,m).
* @since 2017-03-30
* @author R. J. Mathar
*/
template <typename T> std::vector< std::vector<int> > Nice::Alg::subSets(T N, T m)
{
        std::vector<int> fxd ;
        return subSets(N,m,fxd,(T)0) ;
} /* subSets */

/*!******************************************
* @brief construct subsets of size m out of n elements
*
* @param N The number of elements to coose from.
* @param m The number of elements selected in each element of the result.
* @param[in] fxd The sublist of indices which are already fixed.
* @param[in] minm A lower bound of the indices to be taken freely.
* @return ordered lists of numbers in the range 0..n-1 where
* @since 2017-03-30
* @author R. J. Mathar
*/
template <typename T> std::vector< std::vector<int> > Nice::Alg::subSets(T N, T m,
        std::vector<int> fxd, T minm)
{
        std::vector< std::vector<int> > listofm ;
        if ( (int)fxd.size() < m)
        {
                /* still some slots to be filled; attach numbers c to new
                * partial lists of selected elements
                */
                for(T c = minm ; c < N ; c++)
                {
                        std::vector<int> nxtl = fxd ;
                        nxtl.push_back((int)c) ;
                        /* recursive call, one more element fixed now
                        */
                        std::vector< std::vector<int> > sols = subSets(N,m, nxtl,(T)(c+1)) ;
                        /* append all solutions to the existing solutions
                        */
                        listofm.insert(listofm.end(), sols.begin(), sols.end()) ;
                }
        }
        else
                /* all elements fixed, so this is the only choice
                */
                listofm.push_back(fxd) ;

        return listofm ;
} /* subSets */

/* explicit template instantiations
*/
template std::vector< std::vector<int> > Nice::Alg::subSets<short>(short N, short m) ;
template std::vector< std::vector<int> > Nice::Alg::subSets<int>(int N, int m) ;
