#!/usr/bin/env python
# -*- coding: utf-8 -*-

from flask import render_template, make_response, request
from flask_restx import Namespace, Resource

from .devices_access_object import RtcObject as DAO

RTC = DAO()


############################
# REST API
############################

api = Namespace('rtc', description='Real Time Control')

@api.route("/centroid")
class Centroid(Resource):

    def get(self):
        return RTC.centroid

    def put(self):
        cx_cy = api.payload
        RTC.centroid = list(map(float, cx_cy))

@api.route("/shape")
class Shape(Resource):

    def delete(self):
        return RTC.flatten()

@api.route("/loop")
class Loop(Resource):

    # def get(self):
    #     return RTC.todos

    def post(self):
        return RTC.close_loop()

    def delete(self):
        return RTC.open_loop()

@api.route("/frequency")
class Frequency(Resource):

    # def get(self):
    #     return RTC.todos

    def put(self):
        freq = api.payload
        return RTC.frequency(freq)


@api.route("/dark")
class Dark(Resource):

    def get(self):
        return RTC.todos_dark

    def put(self):
        name = api.payload
        return RTC.load_dark(name)

    def post(self):
        return RTC.new_dark()

@api.route("/flat")
class Flat(Resource):

    def get(self):
        return RTC.todos_flat

    def put(self):
        name = api.payload
        print(name)
        return RTC.load_flat(name)

    def post(self):
        return RTC.save_flat()

@api.route("/gain")
class Gain(Resource):

    def get(self):
        return RTC.todos_gain

    def put(self):
        name = api.payload
        return RTC.load_gain(name)

    def post(self):
        gain_type, centroid_size = api.payload
        return RTC.new_gain(gain_type, centroid_size)

@api.route("/tiptilt")
class TipTilt(Resource):

    # def get(self):
    #     return RTC.todos

    def put(self):
        modes = api.payload
        if modes == None:
            return RTC.tiptilt( absolute=True )
        else:
            return RTC.tiptilt( list(map(float, modes)) )

@api.route("/teccam")
class Teccam(Resource):

    def post(self):
        return RTC.teccam_http_start()

    def put(self):
        return RTC.teccam_custom_command(api.payload)


@api.route("/plots/")
class PlotListApi(Resource):

    def get(self):
        """Show RTC related data"""        
        data = RTC.data()
        
        return data

    # def post(self):
    #     return RTC.create(api.payload), 201


@api.route("/file")
class PlotListApi(Resource):

    def post(self):
        """Show RTC related data"""
        
        filename = RTC.save()
        
        return filename

    
@api.route("/plots/<int:id>")
class PlotApi(Resource):

    def get(self, id):
        return RTC.read(id)

    # def delete(self, id):
    #     RTC.delete(id)
    #     return "", 204

    def put(self, id):
        return RTC.update(id, api.payload)

@api.route("/samples")
class SamplesApi(Resource):

    def put(self):
        return RTC.cloud_samples(float(api.payload))


############################
# WEB VIEW
############################

RTC.create({"name": "CRED2 image",
            "id":   "cred",
            "info": [
                ("min", "image min"),
                ("max", "image max"),
                ]
            })

RTC.create({"name": "Mirror shape",
            "id":   "dm",
            "info": [
                ("min", "actuator min"),
                ("max", "actuator max"),
                ]
            })

RTC.create({"name": "Mirror shape without flat",
            "id":   "dm2",
            "info": [
                ("min", "actuator min"),
                ("max", "actuator max"),
                ]
            })

RTC.create({"name": "Centroid cloud",
            "id":   "cloud",
            "info": [
                ("min", "histo min counts"),
                ("max", "histo max counts"),
                ("wx", "size x (px)"),
                ("wy", "size y (px)"),
                ]
            })

web = Namespace('rtc', description='Real Time Control')

@web.route("/")
class Test(Resource):

    def get(self):
        data = RTC.todos
        darks = RTC.todos_dark()
        gains = RTC.todos_gain()
        flats = RTC.todos_flat()
        return make_response(render_template("rtc_panel.html", data=data, darks=darks, flats=flats, gains=gains))

@web.route("/controls/")
class RtcControlsListWeb(Resource):

    def get(self):
        darks = RTC.todos_dark()
        gains = RTC.todos_gain()
        flats = RTC.todos_flat()
        return make_response(render_template("rtc_controls.html", darks=darks, flats=flats, gains=gains))

@web.route("/plots/")
class RtcPlotsListWeb(Resource):

    def get(self):
        data = RTC.todos
        return make_response(render_template("rtc_plots.html", data=data))

@web.route("/plots/<id>")
class RtcPlotsWeb(Resource):

    def get(self,id):
        data = [RTC.read(id)]
        return make_response(render_template("rtc_plots.html", data=data))



############################
# Stream
############################
    
from flask import Response
import json
import time



@web.route('/stream')
class Stream(Resource):
    def get(self):
        def inner():

            previous_res = {}            
            
            while True:
                try:
                    time.sleep(0.5)
                    resource = RTC.data()
                    if previous_res != resource:
                        yield 'data: {}\n\n'.format(json.dumps(resource))
                    else:
                        print("Data are the same as before")
                        
                except GeneratorExit:
                    print("Stop streaming rtc data")
                    return
                
        response = inner()
        print("Streaming rtc data...")

        return Response(response, mimetype="text/event-stream")
