#! /usr/bin/python
# $Header: https://oberon.oapd.inaf.it/shins/branches/gigan-virgin_system/templates/sequencer.py 1885 2019-06-10 08:08:14Z depa $

import argparse
import importlib
import sys
#import signal

from templateLoader import TemplateLoader
from util.shinsLogger import logger
from util import functions as ufun
from util import exception


class Sequencer():
    '''
    Manage a sequence of JSON Observation Blocks (OBs)
    '''

    def __init__(self, embedded=True):
        '''Constructor'''
        self.ob = []
        self.ob_file = None
        self.error = []
        self.tpl = None # The running template
        # signal.signal(signal.SIGINT, self.interrupt)
        # self.original_sigint = signal.getsignal(signal.SIGINT)
        self.quitting = False
        self.embedded = embedded  # To manage sys.exit if called alone


    def execute(self):
        """Execute the ob XML file."""
        self.quitting = False

        try:
            logger.info("--------------------------------------------------------------------------")
            logger.info("| {:^70s} |".format("OB"))
            logger.info("| {:^70s} |".format(self.ob_file))
            logger.info("| {:^70s} |".format("STARTED"))
            logger.info("--------------------------------------------------------------------------")
            ob_xml = ufun.parse(self.ob_file)
            NTpls = len(ob_xml)
        except:
            logger.error("File {} not found".format(self.ob_file))
            return

        #################
        # Template loop #
        #################

        logger.info("Found {} template(s) in OB".format(NTpls))

        for i in range(NTpls):

            #################################
            # Check if a quit was requested #
            #################################

            if self.quitting:
                break

            ##############################################
            # Load the n-th template to check its params #
            ##############################################

            logger.info('======================')
            logger.info("=   Template {}/{}   =".format(i, str(NTpls-1)))
            logger.info('======================')

            tpl_loader = TemplateLoader(ob_xml[i])

            logger.debug("Loading the OB")
            tpl_loader.load()
            logger.debug("Loading the OB... DONE")

            logger.info('----------------------')
            logger.info('Tpl {} : {}'.format(i, tpl_loader.TPLName))
            logger.info('TPLID : {}'.format(tpl_loader.TPLID))
            logger.info('----------------------')

            ###########################
            # Instanciate and call it #
            ###########################

            moduleToUse = importlib.import_module("tpl."+tpl_loader.Script)
            classToUse = getattr(moduleToUse, tpl_loader.Script)
            self.tpl = classToUse()
            self.aborted = False
            
            # Info for fits header
            context = {
                "TPL": tpl_loader.TPLName,
                "TPLSCRIPT": tpl_loader.Script,
                "TPLID": tpl_loader.TPLID,
                "OB": self.ob_file,
            }

            # DPR Keywords
            dpr_type = "DPR_TYPE"
            try:
                context.update({dpr_type: tpl_loader.dpr_type})
            except AttributeError as e:
                logger.warning("SEQUENCER: No DPR TYPE in this template")
            
            status = self.tpl.main(ins_s=tpl_loader.ins_sup,
                                   det_s=tpl_loader.det_sup,
                                   tel_s=tpl_loader.tel_sup,
                                   rtc_s=tpl_loader.rtc_sup,
                                   **context)

            logger.debug("SEQUENCER: {} is returning this: {}".format(__name__, status))
            status = status[0] if type(status) == tuple else status

            if status == 1: # 1 == failure # uvars.EXIT_FAILURE:
                errMsg = "SEQUENCER: Template {} ({}-th template in OB) failed execution".format(tpl_loader.TPLID, i)
                logger.error(errMsg)

            else:
                logger.debug("Template {:s} execution success".format(tpl_loader.TPLID))

        ##########
        # Finish #
        ##########

        logger.info("--------------------------------------------------------------------------")
        logger.info("| {:^70s} |".format("OB"))
        logger.info("| {:^70s} |".format(self.ob_file))
        logger.info("| {:^70s} |".format("FINISHED"))
        logger.info("--------------------------------------------------------------------------")


    def interrupt(self, signal, frame):
        '''Intercept a CTRL+C instead of raising a KeyboardInterrupt
        exception in order to be able to operate on the template class, for example to
        modify a pause attribute or call a method.

        '''

        #signal.signal(signal.SIGINT, self.original_sigint)

        try:
            answer = raw_input("(P)ause, (R)esume, (T)ry again template, (Q)uit:\n")

            if answer.lower().startswith('p'):
                self.pause()
            if answer.lower().startswith('r'):
                self.resume()
            # if answer.lower().startswith('t'):
            #     self.repeat()
            # if answer.lower().startswith('n'):
            #     self.abort()
            if answer.lower().startswith('q'):
                self.quit()
            else:
                pass

        except (KeyboardInterrupt, RuntimeError):
            '''RuntimeError is catched because the
            original_sigint strategy does not work'''

            self.quit()

        except EOFError as e:
            '''On the web server, CTRL+C calls the interrupt. Then, to exit the
            web sever, a CTRL+D have to be intercepted '''

            msg = "SEQUENCER: CTRL-D detected! Exiting for real {} in ob {}".format(self.tpl.name, self.ob_file)
            logger.critical(msg)
            sys.exit(1)
            return

    def pause(self):
        '''
        Put the pause attribute in the running template to True.
        It will be checked in a template method.
        '''

        if self.tpl:
            logger.warning("SEQUENCER: pause called on template {} in ob {}".format(self.tpl.name, self.ob_file))
            logger.warning("SEQUENCER: Template will be paused once it will reach the pause control")
            self.tpl.paused = True


    def resume(self):
        '''
        Restore the pause attribute in the template class to its
        original state. It will be checked in a template method.
        '''

        if self.tpl:
            logger.warning("SEQUENCER: resume called on template {} in ob {}".format(self.tpl.name, self.ob_file))
            self.tpl.paused = False


    def repeat(self):
        '''
        Restore the pause attribute in the template class to its
        original state, BUT go one step behind in the paragraph list.
        So, on resume, it will be repeated.
        It will be checked in a template method.
        '''

        if self.tpl:
            logger.warning("SEQUENCER: repeat called on template {} in ob {}".format(self.tpl.name, self.ob_file))

            raise exception.ManualRetry("Try again")


    def abort(self):
        '''
        Abort the running template. The sequencer will pass to the
        follwing one, if any.
        '''

        if self.tpl:
            msg = "SEQUENCER: aborting template {} in ob {}".format(self.tpl.name, self.ob_file)
            logger.error(msg)
            self.error.append(msg)
            self.tpl.aborted = True
            self.tpl.abort()


    def quit(self):
        '''Abort the execution of the whole OB'''

        msg = "SEQUENCER: Aborting the OB {} and quit the sequencer".format(self.ob_file)
        logger.error(msg)
        self.error.append(msg)
        self.quitting = True # Flask doesn't take sys.exit??

        try:
            self.abort()
            sys.exit(1)
            # self.tpl.obsController.emergencyAbortAll()
            # self.tpl.close_ice()
            # self.tpl.aborted = True

            
        except Exception as e:
            msg = "SEQUENCER: Error trying to abort {}. Exiting anyway.".format(self.ob_file)
            logger.error(msg)
            logger.error(e)
            self.error.append(msg)
            self.error.append(e)

        if not self.embedded:
            sys.exit(1)


#####################
# To call as script #
#####################

def parser():
    """Parser for the sequencer arguments."""

    parser = argparse.ArgumentParser(description='''Sequencer for the SHARK-NIR instrument. It executes Observation Blocks (OBs)
        in the form of XML files calling one or more templates''')

    obname = ['-o', '--ob']
    kwobname = {'type':str, 'help':'OB XML file name'}
    guiname = ['-g', '--gui']

    parser.add_argument(*obname, **kwobname)
    parser.add_argument(*guiname, action='store_true')
    args = parser.parse_args()

    return(args)


def main(args):

    # Have to import signal here due to Flask bug
    import signal

    sequencer = Sequencer(embedded=False)

    # Calling here instead of in Sequencer constructor due to Flask bug
    signal.signal(signal.SIGINT, sequencer.interrupt)
    # sequencer.original_sigint = signal.getsignal(signal.SIGINT)

    logger.info('{}: begin'.format(__name__))

    if args.ob:

        sequencer.ob_file = args.ob
        status = sequencer.execute()
        sys.exit(status)

    else:
        logger.error("See sequencer.py --help for usage")

    logger.info('{}: end'.format(__name__))



if __name__ == "__main__":
    args = parser()
    main(args)
