
from shinsapp import SHINSApp
from util.shinsLogger import logger
from util import variables as uvars
from util import functions as ufun
from util import exception
from time import sleep
import easysnmp
import numpy as np
from astropy.io import fits

class SHARKNIR_gen_base(SHINSApp):
    
    @exception.handler_for_run
    def run(self, ins_s=None, det_s=None, tel_s=None, rtc_s=None, **kwargs):
        '''
        Setup + Image optimizing time
        '''
        
        logger.info(__name__)
        obsController = self.obsController
        self.status = uvars.EXIT_SUCCESS

        fitsDir = uvars.fitsDir + obsController.getCurrentDataDirectory() +'/'

        ###################################
        # Starting instrument setup, is any
        if ins_s:
            
            #########################
            # SHOW INSTRUMENT SETUP #
            #########################
            
            logger.info("Instrument Setup")
            for element, value in ins_s.items():
                logger.debug("{:>22} : {:<22}".format(element, str(value)))
        
            ###############
            # LAMPS SETUP #
            ###############
            if self.paragraph("LAMPS SETUP"): 
                
                elems = ["CAL_FF_LAMP", "CAL_FIBER_FOCUS_LAMP",
                         "CAL_FIBER_DEFOCUS_LAMP"]
                
                for elem in elems:
                    if elem in ins_s:
                        state = ins_s[elem]
                        logger.info("Setup {} to {}".format(elem,state))
                        try:
                            res  = ufun.set_pdu_elem_state(elem, state, pdu=uvars.ipPDU1, aliasDict=uvars.aliasDict)
                            self.status = uvars.EXIT_SUCCESS if res else uvars.EXIT_FAILURE            
                        except easysnmp.EasySNMPTimeoutError as e:
                            logger.error("Failed setting lamp {}: {}".format(elem, e))
                            self.status = uvars.EXIT_FAILURE
                
            ##########################
            # START INSTRUMENT SETUP #
            ##########################
            if self.paragraph("START INSTRUMENT SETUP"): 
                        
                logger.info("Instrument setup starting")
                async = obsController.begin_setupInstrument(ins_s)
                logger.info("Instrument setup started")
                self.status = uvars.EXIT_FAILURE

                
        #################################
        # Starting detector setup, is any
        if det_s:
            
            #####################################
            # CONVERTING DIT INTO SCICAM PARAMS #
            #####################################
            
            if "DIT" in det_s:
                logger.debug("DIT : "+det_s["DIT"]+" specified. Calculating ramp params from this value")
                ramp_s = ufun.calc_ramp_settings(det_s["DIT"], det_s["READOUT"])
                det_s.update(ramp_s)
            
            #######################
            # SHOW DETECTOR SETUP #
            #######################
            
            logger.info("Detector Setup")
            for element, value in det_s.items():
                logger.debug("{:>22} : {:<22}".format(element, str(value)))
                        
            #####################
            # START SASHA SETUP #
            #####################
            if self.paragraph("START SASHA SETUP"): 
                        
                logger.info("Setting up SASHA ...")
                asyncResSashaSetup = obsController.begin_sashaSetup(det_s)
                logger.info("Setting up SASHA ... DONE")                            


        ##################################
        # Waiting instrument setup, is any
        if ins_s:
                
            #########################
            # WAIT INSTRUMENT SETUP #
            #########################
            if self.paragraph("WAIT INSTRUMENT SETUP"): 
                
                logger.info("Waiting Instrument setup to complete...")
                self.status = obsController.end_setupInstrument(async)
                ufun.error_handler(self.status, "setupInstrument")
                logger.info("Instrument setup... DONE")        
                
                
            ######################
            # CHECK LAMPS STATUS #
            ######################
            if self.paragraph("CHECK LAMPS SETUP"): 

                if elems:
                    sleep(2)
                
                lamps_s = {}
                for elem in elems:
                    if elem in ins_s:
                        try:
                            res  = ufun.get_pdu_elem_state(elem, pdu=uvars.ipPDU1, aliasDict=uvars.aliasDict)
                            logger.debug("Lamp {} is now {}".format(elem, res))
                            # Filling the lamps status
                            lamps_s.update({elem: res})
                        except easysnmp.EasySNMPTimeoutError as e:
                            logger.error("Failed getting elem {}: {}".format(elem, e))
                            self.status = uvars.EXIT_SUCCESS if res else uvars.EXIT_FAILURE
            
                # Passing the lamps status to OS
                if lamps_s:
                    self.obsController.setLampsStatus(lamps_s)


        #############################################
        # Waiting detector setup, is any, then expose
        if det_s:
                
            ####################
            # WAIT SASHA SETUP #
            ####################
            if self.paragraph("WAIT SASHA SETUP"): 
                
                logger.info("Waiting SASHA setup ...")
                self.status = obsController.end_sashaSetup(asyncResSashaSetup)
                logger.info("Waiting SASHA setup ... DONE")
                        

                
            #############################################
            # Take the image if kwargs["expose"] is not set to False
            if "expose" in kwargs and kwargs["expose"] == False:
                pass
            else:
                
                ##################
                # EXPOSING SASHA #
                ##################
                if self.paragraph("EXPOSING SASHA"): 
                    
                    logger.info("Waiting camera exposure to complete ...")
                    self.status, self.filenames = obsController.sashaExpose()
                    logger.info("Waiting camera exposure to complete ... DONE")
                    logger.info("FITS files are "+str(self.filenames))
                        
                    
                #####################
                # SHOW IMAGES STATS #
                #####################
                if self.paragraph("IMAGE(S) STATS"): 
                                    
                    for image in self.filenames:
                        fitsDir = uvars.fitsDir + obsController.getCurrentDataDirectory() +'/'
                        data = fits.getdata(fitsDir+image)
                        logger.debug("-------------------------")
                        logger.debug("Image  : {}".format(image))
                        logger.debug("Mean   : {}".format(np.mean(data)))
                        logger.debug("Median : {}".format(np.median(data)))
                        logger.debug("Stddev : {}".format(np.std(data)))
                        logger.debug("Zeroes : {}".format(np.count_nonzero(data==0)))

                    if "NDIT" in det_s:
                        if int(det_s["NDIT"]) != len(self.filenames):
                            msg = "FITS requested: {}, FITS saved: {} !".format(det_s["NDIT"], len(self.filenames))
                            logger.warning(msg)
                            ufun.error_handler(self.status, "sashaExpose")
                        
                    if self.status:
                        ufun.error_handler(self.status, "sashaExpose")

