
from shinsapp import SHINSApp
from util.shinsLogger import logger
from alignment.allineamento import Align_the_coronagraph as aly
from util import variables as uvars
from util import exception
from util import functions as ufun
import time
import numpy as np

from SHARKNIR_gen_base import SHARKNIR_gen_base

class SHARKNIR_gen_tec_PupilLyotAlignment(SHINSApp):

    @exception.handler_for_run
    def run(self, ins_s=None, det_s=None, tel_s=None, rtc_s=None, **kwargs):
        '''template to align Pupil on lyot. Steps:
        . instrument setup START and WAIT
        . setup SASHA
        . expose SASHA
        . retrieve FITS file
        . retrieve apo mask position
        . instrument setup START and WAIT
        . retrieve pupil position and move tip/tilt mirror (iterate until convergence)
        . save residual displacement X,Y
        . instrument setup START and WAIT (for rot alignment)
        . setup SASHA
        . expose SASHA
        . retrieve FITS file
        '''
        logger.info(__name__)

        self.status = uvars.EXIT_SUCCESS
        time_out = 120000        
        obsController = self.obsController
        
        gen_base = SHARKNIR_gen_base()

        fitsDir = uvars.fitsDir + obsController.getCurrentDataDirectory() +'/'

        saly = aly(camera='HAWAII H-2RG')


        ###############################
        # SETUP FOR LYOT CENTER CHECK #
        ###############################

        align_ins_s = {
            "CAL_MIRROR_DEP" : "IN",
            "CAL_FF_LAMP": "ON",
            #"LAMP_TEL_SIM" : "OFF",
            "SCI_FILT_W1" : "ND2",
            "SHUTTER" : "IN",
        }

        ins_s.update(align_ins_s)

        orig_det_s = {
            "NGROUPS":"1",
            "NREADS":"1",
            "NDROPS":"0",
            "NDIT": "1",
            "NCOADDS":"1",
            "READOUT":det_s["READOUT"]
        }

        #################
        # Setup + image #
        #################

        logger.info("Waiting setup to retrive coro mask center ...")
        self.status, self.filenames, _ = gen_base.main(ins_s=ins_s, det_s=orig_det_s, **kwargs)
        logger.info("Waiting setup to retrive coro mask center ... DONE")

        #########################
        # DETERMINE LYOT CENTER #
        #########################
        if self.paragraph("DETERMINE LYOT CENTER"):

            logger.info("Determine LYOT center")
            
            which_lyot = {
                "FQPM_LYOT_STOP": 'FQPM',
                "GAUSS_LYOT_STOP": 'GAUSS',
                "SP_LYOT_STOP": "SP1",
                }
            
            our_lyot = which_lyot[ins_s["LYOT_GRISM_W"]]
            
            saly.AddACoro(our_lyot)
            
            if our_lyot == 'FQPM':
                saly.FQPM.LS.LoadImage(fitsDir+self.filenames[-1])
                # saly.FQPM.LS.removeHotPixels(verbose=True)
                saly.FQPM.LS.Center(method='isophote', level=15000, display=False)
                lyotCX = (saly.FQPM.LS.xc).n
                lyotCY = (saly.FQPM.LS.yc).n
            elif our_lyot == 'GAUSS':
                saly.GAUSS.LS.LoadImage(fitsDir+self.filenames[-1])
                # saly.GAUSS.LS.removeHotPixels(verbose=True)
                saly.GAUSS.LS.Center(method='isophote', level=15000, display=False)
                lyotCX = (saly.GAUSS.LS.xc).n # It's an ufloat(2, 0.25)
                lyotCY = (saly.GAUSS.LS.yc).n # .n is the number, .s is the error
            elif our_lyot == 'SP1':
                saly.SP1.LS.LoadImage(fitsDir+self.filenames[-1])
                # saly.SP1.LS.removeHotPixels(verbose=True)
                saly.SP1.LS.Center(method='isophote', level=15000, display=False)
                lyotCX = (saly.SP1.LS.xc).n
                lyotCY = (saly.SP1.LS.yc).n
            
            logger.warning("CHECK: Lyot center values are {}, {}".format(lyotCX, lyotCY))

        #####################################
        # PREPARE SETUP FOR LBT orientation #
        #####################################

        logger.info("Setup to retrive LBT orientation: move coro masks to hole")

        align_ins_s = {
            "LYOT_GRISM_W": "HOLE",
            "CAL_MIRROR_DEP" : "OUT",
            "CAL_FF_LAMP": "OFF",
            "SCI_FILT_W1" : "HOLE",
            "SHUTTER" : "OUT",
        }

        #################
        # Setup + image #
        #################

        logger.info("Waiting setup to retrive coro mask center ...")
        self.status, self.filenames, _ = gen_base.main(ins_s=align_ins_s, det_s=det_s, **kwargs)
        logger.info("Waiting setup to retrive coro mask center ... DONE")

        ##############################
        # ITERATIONS TO ALIGN PUPILS #
        ##############################

        # this loop goes up to convergence or to a max num of
        # iteration, after which it fails

        converged = False
        thrsConv = 3.0 # pixels.
        nIteration = 0
        maxIteration = 5

        LBTCXarr = []
        LBTCYarr = []

        while nIteration < maxIteration and not converged:
            
            if self.paragraph("ITERATION {} of {}".format(nIteration, maxIteration)):

                if nIteration > 0:
                
                    ##################
                    # EXPOSING SASHA #
                    ##################
                
                    logger.info("Waiting camera exposure to complete ...")
                    self.status, self.filenames = obsController.sashaExpose()
                    logger.info("FITS files are "+str(self.filenames) )
                
                else:
                    logger.debug("use last image of LBT taken")
                    pass
                
                ##############################
                # DETERMINE LBT PUPIL CENTER #
                ##############################
                
                logger.info("Determine LBT center")
                
                saly.Pup.LoadImage(fitsDir+self.filenames[-1])
                # saly.Pup.removeHotPixels(verbose=True)
                
                # saly.Pup.LoadTemplate("SCICAM")
                # saly.Pup.ZumTemplate(shift=(-3500,-1700),scale=4.0)
	        # saly.PupilCenter(method='FLANN', display=False, distance_ratio=0.5)
                saly.PupilCenter(method='Hough', rmin=335, rmax=345, hthresh=(100,150), display=False)
                # LBTCX, LBTCY = saly.Pup.xc, saly.Pup.yc
                LBTCX, LBTCY = (saly.Pup.xc).n, (saly.Pup.yc).n
                
                logger.warning("CHECK: LBT center are {}, {}".format(LBTCX, LBTCY))
                
                ####################
                # GET CURRENT DROT #
                ####################
                
                inDict = {"DEROT":-9999.9}
                unit = "DEGREE"
                
                self.status, outDict = obsController.getPosition(unit, inDict)
                logger.warning("DEROT is: {} degrees".format(outDict))
                
                ########################
                # COMPUTE DISPLACEMENT #
                ########################
                
                logger.info("Calculate displacement between orientation and lyot centers")
                
                deltaX =  LBTCX - lyotCX
                deltaY =  LBTCY - lyotCY
                
                logger.info("Deltas are {} {}".format(deltaX, deltaY))
                
                thetaInTT = ufun.transform_offset(disp=[deltaX, deltaY],
                                                  thetaRot=outDict["DEROT"])
                
                logger.info("Apply this displacementon INBEAM_TT: {}".format(thetaInTT))
                
                #########################
                # GET CURRENT INBEAM_TT #
                #########################
                
                logger.info("Take current INBEAMTT positions")
                
                outDict = {}
                
                self.status, ttx = obsController.getPosition("STEPS", {"INBEAM_TT_X": -9999})
                self.status, tty = obsController.getPosition("STEPS", {"INBEAM_TT_Y": -9999})
                
                outDict.update(ttx)
                outDict.update(tty)
                
                logger.warning("INBEAM_TT are: {} steps".format(outDict))
                
                tt_ins_rel = {
                    "INBEAM_TT_X": thetaInTT[0],
                    "INBEAM_TT_Y": thetaInTT[1]
                }
                
                ######################
                # CHECK IF CONVERGED #
                ######################
                
                logger.debug("CHECK: Displacement: deltaX: {} deltaY: {}".format(deltaX, deltaY))
                
                converged = True if ( (np.abs(deltaX) <= thrsConv) and (np.abs(deltaY) <= thrsConv) ) else False
                
                logger.debug("CHECK: Converged: {}".format(converged))
                
                nIteration += 1
                
                if not converged:
                
                    ##################################
                    # SUM AND APPLY THE DISPLACEMENT #
                    ##################################
                
                    logger.info("Sum values")
                
                    tt_ins_n = {k: outDict.get(k, 0) + tt_ins_rel.get(k, 0) for k in set(outDict) & set(tt_ins_rel)}
                    keys_values = tt_ins_n.items()
                    tt_ins_s = {k:str(v) for (k,v) in keys_values}
                
                    logger.info("Waiting setup to apply displacement...")
                    gen_base.main(ins_s=tt_ins_s, **kwargs)
                    logger.info("Waiting setup to apply displacement... DONE")

        ###################
        # END OF THE LOOP #
        ###################
        if self.paragraph("END OF THE LOOP"):

            LBTCXarr.append(lyotCX)
            LBTCYarr.append(lyotCY)
            
            print(LBTCXarr)
            print(LBTCYarr)
            
            logger.warning("CHECK: LBT pupil center: [{:.2f}, {:.2f}]".format(saly.Pup.xc, saly.Pup.yc))
            logger.warning("CHECK: Lyot center: [{:.2f}, {:.2f}]".format(lyotCX, lyotCY))
            logger.warning("CHECK: Displacement: deltaX: {} deltaY: {}".format(deltaX, deltaY))
            
            ##################################
            # SAVE RESIDUAL DISPLACEMENT X,Y #
            ##################################
            # .... to HDF5 file
            # logger.warning("Saving residual displacement X, Y on HDF5 file NOT YET IMPLEMENTED")
            
            # if not converged:
            #     logger.error("Pupil alignment did not converged")
            #     self.status = uvars.EXIT_FAILURE
            #     raise RuntimeError("Pupil alignment did not converged")
