
import os
import time
import numpy as np
from lxml import etree
import easysnmp
from itertools import cycle
from time import sleep
from util import variables as uvars
from util.shinsLogger import logger
from util import exception
from astropy.coordinates import SkyCoord
from astropy.time import Time
import astropy.units as u


###############
# Miscellanea #
###############

def calc_ramp_settings(dit, readout, ngmin=2, ngmax=None, keep_ngmax=False, diff_threshold=0.05):
    """ Determine ramp settings for SCICAM acquisitions

    This script proposes a strategy to set SCICAM acquisition parameters given an input
    targeted exposure time and a given setting of the camera (type of window and clock time).
    Created on Mon Aug 16 18:46:58 2021
    @author: DV


    This function computes the SCICAM ramp settings (NGROUPS, NDROPS) that best match
    a targeted integration time. NREADS is always 1.

    Parameters
    ==========
    dit            : the targeted integration time in seconds
    ftime_ms       : the frame time in milliseconds (readout mode dependent)
    ngmin          : the minimum number of groups
    ngmax          : the maximum number of groups
    keep_ngmax     : set True if you want to force the ramp to have exactly ngmax groups. If False, ngroups
                     is optimized so to get as close as possible to the targeted integration time.
    diff_threshold : difference ratio between dit and calculated. 0.05 = 5%
    """

    dit = float(dit)*1000 # transform to milliseconds

    frame_type = {
        "128x128": 71.76,
        "coro_stripe": 128.322,
        "256x256": 274.5,
        "Center": 2675.1,
        "Full_Image": 4278.3,
    }

    ftime_ms = frame_type[readout]

    # Case 0. Single frame
    if dit < ftime_ms:
        num_groups = 1
        num_reads  = 1
        num_drops  = 0

        return { "NGROUPS": str(num_groups),
                 "NREADS": str(num_reads),
                 "NDROPS": str(num_drops),
                }

    # Total number of frames requested to match integration time
    nftot = int(np.ceil(dit/ftime_ms))

    if ngmax is None:
        ngmax=nftot

    # Case 1.
    # Requested frames are more than groups. Need to insert drop frames into each group.
    # Optimize ndrops AND NGROUPS to get close to the requested itime
    if nftot > ngmax:
        num_reads = 1
        if keep_ngmax:
            ngmin = ngmax
        #else:
        #    ngmin = ngmax - 1 if ngmax <= 3 else 2

        nf_diff_prev = 50.0
        frac_prev = 1.0

        for i in range(ngmax, ngmin-1, -1):
            num_groups = i
            num_drops  = int(np.rint((dit - num_groups * num_reads * ftime_ms) / (num_groups - 1) / ftime_ms))
            nftot_res = num_groups * num_reads + (num_groups - 1) * num_drops

            # Break conditions
            nf_diff = nftot_res - (dit / ftime_ms)
            frac = nf_diff * ftime_ms / dit

            # This keeps the highest ngroups if same values for multiple combinations
            if (nf_diff < nf_diff_prev or frac < frac_prev):
                ng_best = num_groups
                nd_best = num_drops
                nf_diff_prev = nf_diff
                frac_prev = frac

            if nf_diff <= 1 or frac <= diff_threshold:
                break

        num_groups = ng_best;
        num_drops = nd_best;

        '''
        num_groups = ngmax
        num_reads  = 1
        num_drops  = (ExpTime - num_groups*FrameTime) / (num_groups - 1) / FrameTime
        num_drops_array = np.array([np.floor(num_drops), np.ceil(num_drops)])
        EffExpTime = FrameTime * (num_groups*num_reads + (num_groups - 1)*num_drops_array)
        Delta = np.abs(ExpTime-EffExpTime)
        num_drops = int(num_drops_array[np.argmin(Delta)])
        '''

    # Case 2.
    # Requested frames are less than groups. Save all of them in separated groups.
    else:
        num_groups = nftot
        num_reads  = 1
        num_drops = 0

    EffExpTime = ftime_ms * (num_groups*num_reads + (num_groups-1)*num_drops)
    logger.debug("setINDI sasha.num_groups.value={}".format(num_groups))
    logger.debug("setINDI sasha.num_reads.value={}".format(num_reads))
    logger.debug("setINDI sasha.num_drops.value={}".format(num_drops))
    logger.debug("Effective exposure time [ms]: {:.2f}".format(EffExpTime))
    logger.debug("Delta [ms]: {:.2f}".format(dit-EffExpTime))


    return { "NGROUPS": str(num_groups),
             "NREADS": str(num_reads),
             "NDROPS": str(num_drops),
            }


def make_spiral(step, repeat):

    def move_right(x,y,step):
        return [x+step, y]

    def move_down(x,y,step):
        return [x, y-step]

    def move_left(x,y,step):
        return [x-step, y]

    def move_up(x,y,step):
        return [x, y+step]

    moves = [move_right, move_down, move_left, move_up]

    def gen_points(step, repeat):
        _moves = cycle(moves)
        n = 1
        pos = [0,0]
        times_to_move = 1

        yield pos # initial

        while True:
            for _ in range(2):
                move = next(_moves)
                for _ in range(times_to_move):
                    if n >= repeat:
                        return
                    pos = move(0,0,step) #move(*pos)
                    n+=1
                    yield pos

            times_to_move+=1

    spiral = gen_points(step, repeat)
    return list(spiral)

            
###########################
# Lamps-related functions #
###########################


def switch_on_off_pdu_elements(pdu, aliasDict, onoff="OFF"):

    elems = aliasDict

    # LAMPS SETUP #

    for elem in elems:
        state = onoff
        logger.info("Setup {} to {}".format(elem,state))
        try:
            res  = set_pdu_elem_state(elem, state, pdu=pdu, aliasDict=aliasDict)
            status = uvars.EXIT_SUCCESS if res else uvars.EXIT_FAILURE

        except easysnmp.EasySNMPTimeoutError as e:
            logger.error("Failed setting lamp {}: {}".format(elem, e))
            status = uvars.EXIT_FAILURE

    sleep(1)

    # LAMPS STATUS #

    for elem in elems:
        try:
            res  = get_pdu_elem_state(elem, pdu=pdu, aliasDict=aliasDict)
            logger.debug("Lamp {} is now {}".format(elem, res))
        except easysnmp.EasySNMPTimeoutError as e:
            logger.error("Failed getting elem {}: {}".format(elem, e))
            status = uvars.EXIT_FAILURE


def get_pdu_elem_state(elem, getattribute="OutletState",
                       pdu=uvars.ipPDU1, aliasDict=uvars.aliasDict):

    session = easysnmp.Session(hostname=pdu,
                      security_username=uvars.pduUID,
                      version=uvars.snmpVer)
    state = uvars.dictOutletControl[getattribute]
    device = aliasDict[elem]
    res = session.get(state.format(device))
    return uvars.lampStateDict[int(res.value)]


def set_pdu_elem_state(elem, state, setattribute="OutletCommand", pdu=uvars.ipPDU1,
                       aliasDict=uvars.aliasDict):

    session = easysnmp.Session(hostname=pdu,
                      security_username=uvars.pduUID,
                      version=uvars.snmpVer)
    send = uvars.dictOutletControl[setattribute]
    device = aliasDict[elem]
    command = uvars.outletCmdDict[state]
    res = session.set(send.format(device), command, snmp_type='INTEGER')
    return res


#########################
# XML-related functions #
#########################


def dict_to_xml(tag, d):
    """Transform a dictionary into an xml. Params:
    tag: root tag name, and
    d: dict
    """

    element = etree.Element(tag)
    for key, val in d.iteritems():
        child = etree.Element(key)
        child.text = str(val)
        element.append(child)
    return element


def parse(xml_file):
    """Reads the XML file. It can be a ref setup, or an OB"""

    parser = etree.XMLParser(remove_comments=True)
    tree = etree.parse(xml_file, parser=parser)

    logger.info("Ok parsing {}".format(xml_file))

    valid = tree.getroot()

    return valid

########################
# Conversion functions #
########################

def transform_offset(disp=[0, 0], thetaRot=0.0, conv="Pup"):

    if conv == "Pup":
        thetaRef = np.deg2rad(95.0) # bearing angle
        intMat = [[ 0.0010, 0.0016],
                  [-0.0016, 0.0015]]

    elif conv == "PSF":
        thetaRef = np.deg2rad(0.0)
        intMat = [[-11.092,  0.0   ],
                  [ 0.0,    -10.945]]
        
    elif conv == "SOUL":
        thetaRef = np.deg2rad(95.0) # bearing angle
        intMat = [[ 110.268,  -37.687 ],
                  [-36.704,   -108.3 ]]
        
    thetaRot = np.deg2rad(thetaRot)

    thetaDiff = thetaRot - thetaRef

    R = np.asarray([[np.cos(thetaDiff), -np.sin(thetaDiff)],
                    [np.sin(thetaDiff),  np.cos(thetaDiff)]])

    RI = np.dot(R, intMat)

    invRI = np.linalg.inv(RI)

    theta = np.dot(invRI, np.asarray(disp)) 

    return theta


    


def to_teccam(x, y):
    '''
    Convert SCICAM coordinates [px],
    into TECCAM coordinates, which are scaled from -1 to 1.
    0.47px on SCICAM corresponds to 0.01 on TECCAM.
    '''

    conversion = (0.47/0.01)

    return [+x/conversion, -y/conversion]


def guess(value):
    """Guess the type of a variable"""
    
    try:
        val = int(value)
        return val
    except:
        try:
            #float
            val = float(value)
            return val
        except:
            try:
                # list or dict
                val = json.loads(value)
                return val
            except:
                # string
                return value

##########################
# NCPA-related functions #
##########################

def label_coefficients(coefficients):
    labels = ['AstO','AstY','ComY','ComX','TriY','TriO','Sph','AstYII','AstOII','TetY','TetO',
              'ComXII','ComYII','TriOII','TriYII','PenY','PenX','SphII', 'AstOIII', 'AstYIII',
              'TetOII', 'TetYII', 'EsaY', 'EsaX', 'ComYIII', 'ComXIII', 'TriYIII', 'TriOIII',
              'PentaXII', 'PentaYII', 'EptaY', 'EptaX']
    
    l = len(labels)
    c = len(coefficients)

    if c<=l:
        labels = labels[0:c]
    else:
        modes = ['Mode' + str(m) for m in range(l, c)]
        labels = labels + modes

    return list(zip(labels, coefficients))


def ncpa_masterize(fitsDir, filename, master_array):
    '''
    Apply  the CDS techique on the image and mean it over the number of images.
    Returns the master_array with 6 images.
    '''

    from alignment.allineamento import read_image
    from alignment.utils import find_hot_pixels

    img_arr = []
    for image in filename:
        img = read_image(fitsDir+image, header=False, stack='', crop=1)
        img_minus_dark = img[-1] - img[0] # always 2 slices as NGROUPS=2 in .ref
        img_arr.append(img_minus_dark)

    master_img = np.mean(np.array(img_arr), axis=0)
    master_img,_ = find_hot_pixels(master_img, num_iter=3) 

    master_array.append(master_img)

    return master_array


def call_onera(master_array, occultation=0.1):
    from idlpy import *

    logger.debug("Prepare defocuses")

    # Configured to use all 6 images in master_array
    #          -----focus fiber----    -------defocus fiber -----
    #          hole  pd_neg  pd_pos    hole     pd_neg  pd_pos
    defocuses = [0, -26.378, +26.378, -13.189, -39.568,  +13.189 ]
    n_images = len(defocuses)
    
    # Sort images: focus HOLE3 first and then the others ins ascending order of defocus 
    master_array[1:] = [a for _,a in sorted(zip(defocuses[1:],master_array[1:]))]
    defocuses = [0, -39.568, -26.378, -13.189, +13.189, +26.378]
    
    # # TEST DANIELE : pop out the focus HOLE3 image.
    # master_array = master_array[1:]
    # defocuses = defocuses[1: n_images] 
    # n_images = len(defocuses)


    # Parameters for ONERA
    logger.debug("Declare parameters in IDL")
    IDL.master_array = np.array(master_array)
    IDL.Delta_z = np.array(defocuses)
    IDL.Fnum = 32.1
    IDL.wavelength = 1.593 # Associated to SCI_FILT_W1=NB_H2
    IDL.defoc_array_a4=np.tile([+0.,+0.,+0.],(n_images,1))
    IDL.nModes = 36
    IDL.opt = 'full'
    IDL.windou = np.array([0,0,128])
    IDL.pup_mask = 0
    IDL.visu = 0 # don't display
    IDL.occultation = occultation
    
    # Calling Daniele's run_pd IDL routine that calls ONERA code.
    #logger.debug("Compliling wrapper run_pd")
    #IDL.run("cd, '{}'".format("/usr/local/harris/idl87/lib/PD_IDL"))
    #IDL.run(".compile run_pd")
    
    logger.debug("Calling wrapper run_pd")
    IDL.run("run_pd, master_array, Delta_z, Fnum, wavelength, nModes, opt, defoc_array_a4, windou=windou, pup_mask=pup_mask, Zernikes=Zernikes, Niter=Niter, diversities=diversities, pup_diam=pup_diam, oversample=oversample, visu=visu, occultation=occultation")


    # Print outputs on the screen
    np.set_printoptions(precision=2,suppress=True)
    logger.debug("\nNumber of iterations to reach convergence: {}".format(IDL.Niter))
    
    logger.debug("\nDiversities in rad rms (first mode is tip):")
    logger.debug(IDL.diversities)
    
    logger.debug("\nTip/Tilt in pixels:")
    logger.debug(4/np.pi*IDL.oversample*IDL.diversities[:,0:2])    
    
    logger.debug("Enumerating coefficients in nm:")
    zernikes = IDL.Zernikes[1:]*1e3*IDL.wavelength/2./np.pi #[nm]
    zernikes_names = label_coefficients(zernikes)
    
    for element, value in zernikes_names:
        logger.debug("Element : % 8s  -  Value %6.2f    " % (element, value))
        
    logger.debug("\nTotal WFE in nm: %6.2f" %(np.sqrt(np.sum(zernikes**2))))

    print([round(value, 2) for (element, value) in zernikes_names])

#################################
# Coordinates-related functions #
#################################

    
def to_radians(tel_s):
    """Marco version for coordinate to radian conversion"""
    
    # print(tel_s)

    logger.info("Converting coordinates to radian")

    tel_s["Epoch"] = str(Time.now().jyear)
    
    if ('OBJRA' in tel_s) and ('OBJDEC' in tel_s):
        obsCoord = SkyCoord(tel_s['OBJRA'], tel_s['OBJDEC'], unit=("hourangle", "deg"))
        tel_s['OBJRA'] = '{}'.format(obsCoord.ra.radian)
        tel_s['OBJDEC'] = '{}'.format(obsCoord.dec.radian)

    if ('PMRA' in tel_s) and ('PMDEC' in tel_s):
        obsCoord = SkyCoord(tel_s['PMRA'], tel_s['PMDEC'], unit=("mas", "mas"))
        tel_s['PMRA'] = '{}'.format(obsCoord.ra.wrap_at(np.pi*u.rad).radian ) 
        tel_s['PMDEC'] = '{}'.format(obsCoord.dec.wrap_at(np.pi*u.rad).radian)

    if ('GSRA' in tel_s) and ('GSDEC' in tel_s):
        obsCoord = SkyCoord(tel_s['GSRA'], tel_s['GSDEC'], unit=(u.hourangle, u.degree))
        tel_s['GSRA'] = '{}'.format(obsCoord.ra.radian)
        tel_s['GSDEC'] = '{}'.format(obsCoord.dec.radian)

    if ('PMGSRA' in tel_s) and ('PMGSDEC' in tel_s):
        obsCoord = SkyCoord(tel_s['PMGSRA'], tel_s['PMGSDEC'], unit=("mas", "mas"))
        tel_s['PMGSRA'] = '{}'.format(obsCoord.ra.wrap_at(np.pi*u.rad).radian ) 
        tel_s['PMGSDEC'] = '{}'.format(obsCoord.dec.wrap_at(np.pi*u.rad).radian)
                
    if ('AORA' in tel_s) and ('AODEC' in tel_s):
        obsCoord = SkyCoord(tel_s['AORA'], tel_s['AODEC'], unit=(u.hourangle, u.degree))
        tel_s['AORA'] = '{}'.format(obsCoord.ra.radian)
        tel_s['AODEC'] = '{}'.format(obsCoord.dec.radian)

    if ('PMAORA' in tel_s) and ('PMAODEC' in tel_s):
        obsCoord = SkyCoord(tel_s['PMAORA'], tel_s['PMAODEC'], unit=("mas", "mas"))
        tel_s['PMAORA'] = '{}'.format(obsCoord.ra.wrap_at(np.pi*u.rad).radian ) 
        tel_s['PMAODEC'] = '{}'.format(obsCoord.dec.wrap_at(np.pi*u.rad).radian)

    if ('OFFSETX' in tel_s) and ('OFFSETY' in tel_s):
        obsCoord = SkyCoord(tel_s['OFFSETX'], tel_s['OFFSETY'], unit=("arcsec", "arcsec"))
        tel_s['OFFSETX'] = '{}'.format(obsCoord.ra.wrap_at(np.pi*u.rad).radian ) 
        tel_s['OFFSETY'] = '{}'.format(obsCoord.dec.wrap_at(np.pi*u.rad).radian)
    
    return tel_s


def error_handler(status, method=""):
    if status == uvars.EXIT_FAILURE:
        msg = "{:s} returned EXIT_FAILURE".format(method)
        raise exception.TemplateException(msg)
